﻿local addon, ns = ...

local MAJOR, MINOR = "gCore-3.0", 2
local gCore, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if (not gCore) then return end 

-- Lua API
local _G = _G
local pcall = pcall
local print = print 
local tinsert, tremove = table.insert, table.remove
local pairs, select = pairs, select
local type = type

local SetCVars
local GetBinary, GetBoolean
local ValidateTable, CleanTable, Embed

gCore._CALLBACKFRAME = gCore._CALLBACKFRAME or CreateFrame("Frame", "gCore30CallbackFrame")
gCore._FRAME = gCore._FRAME or CreateFrame("Frame", "gCore30EventFrame")

--gCore._ADDON = gCore._ADDON or {}
gCore._ADDONLOADED = gCore._ADDONLOADED or nil
gCore._ADDONS = gCore._ADDONS or {}

gCore._MODULE = gCore._MODULE or {}
gCore._MODULES = gCore._MODULES or {}

gCore._CALLBACKS = gCore._CALLBACKS or {}
gCore._EVENTS = gCore._EVENTS or {}
gCore._EVENTSTATUS = gCore._EVENTSTATUS or {}
gCore._TASKS = gCore._TASKS or {}

gCore._SLASH = gCore._SLASH or {}

--_ADDON = gCore._ADDON
_ADDONS = gCore._ADDONS
_CALLBACKFRAME = gCore._CALLBACKFRAME
_CALLBACKS = gCore._CALLBACKS
_EVENTS = gCore._EVENTS
_EVENTSTATUS = gCore._EVENTSTATUS
_FRAME = gCore._FRAME
_MODULE = gCore._MODULE
_MODULES = gCore._MODULES
_SLASH = gCore._SLASH
_TASKS = gCore._TASKS

_RESERVEDEVENTS = { 
	["ADDON_LOADED"] 			= true; 
	["PLAYER_LOGIN"] 			= true; 
	["PLAYER_LOGOUT"] 			= true;
	["PLAYER_ENTERING_WORLD"] 	= true;
}

for i,v in pairs(_RESERVEDEVENTS) do
	_EVENTSTATUS[i] = _EVENTSTATUS[i] or {}
end

-- no operation; might as well make it a global
noop = function() return end

--
-- A little internal error handling mainly to prevent the module
-- from breaking from errors during the startup process
--	
-- We mainly use this for functions connected to reserved events
--
local safecall
do
	local print = function(...)
		_G.print((GetAddOnMetadata(addon, "Title") or addon) .. " encountered an error:")

		for i = 1, select("#", ...) do
			_G.print("|cffff8800" .. (select(i, ...)) .. "|r")
		end
	end

	safecall = function(func, ...)
		-- just a little fail-safe to avoid breaking the fail-safe...
		if type(func) == "function" then
			local ret = { pcall(func, ...) }
			
			if ret[1] then 
				if #ret > 1 then
					return select(2, ret)
				end
			else
				-- fire a non-breaking error to the client, 
				-- but continue execution as otherwise normal
				geterrorhandler()(ret[2], 2)
				
				-- also print the error massage to the chat,
				-- mainly because I want people to report these errors!
				print(ret[2])
				
				-- shouldn't be any return values here, but you never know
				if #ret > 2 then
					return select(3, ret)
				end
			end
		end
	end
end

------------------------------------------------------------------------------------------------------------
-- 	Main Event Frame
------------------------------------------------------------------------------------------------------------
for event, isReserved in pairs(_RESERVEDEVENTS) do
	if (isReserved) then 
		_FRAME:RegisterEvent(event)
	end
end

local OnUpdate = function(self, elapsed)
	local run, kill
	for moduleName, tasks in pairs(_TASKS) do
		for reference, task in pairs(tasks) do
			
			task.elapsed = (task.elapsed or 0) + elapsed
			
			run, kill = true, nil

			-- is there a maximum duration, and has it been exceded?
			if (task.duration) then
				task.total = (task.total or 0) + elapsed
				
				-- its been exceded, so cancel the run, and kill off the timer
				if (task.total > task.duration) then
					kill = true
					run = nil
				end
			end

			-- is there a delay before the initial run?
			if (task.delay) then
				if (task.delay > task.elapsed) then
					run = nil
				else
					task.delay = nil
				end
			end

			-- is this a repeating timer?
			if (task.interval) then
				if (task.elapsed >= task.interval) then
					task.elapsed = task.elapsed - task.interval
				else
					run = nil
				end
			end

			-- shall it run, or was it a reason not to?
			if (run) then
				task.func(_MODULES[moduleName])

				-- if this isn't a repeating timer, we terminate it after this initial run
				if (not task.interval) then 
					kill = true
				end
			end
			
			-- the timer will die!
			if (kill) then
				_TASKS[moduleName][reference] = nil
			end
		end
	end
end

local OnEvent = function(self, event, ...)
	if (_RESERVEDEVENTS[event]) then
		_FRAME[event](self, event, ...)
		return
	end
	
	if (_EVENTS[event]) then
		for module, func in pairs(_EVENTS[event]) do
			if (type(func) == "string") then
				if (_MODULES[module][event]) then
					_MODULES[module][event](_MODULES[module], ...)
				end
				
			elseif (type(func) == "function") then
				func(_MODULES[module], event, ...)
			end
		end
	end
end

_FRAME.ADDON_LOADED = function(self, event, name, ...)
	if (name == addon) then
		gCore._ADDONLOADED = true
	end
	
	-- calling the pre-initialization for ALL modules first
	for name, module in pairs(_MODULES) do
		if not(_EVENTSTATUS[event][module:GetName()]) then
			safecall(module.Init, module, ...)
		end
	end

	-- calling the initialization functions 
	for name, module in pairs(_MODULES) do
		if not(_EVENTSTATUS[event][module:GetName()]) then
			safecall(module.OnInit, module, ...)
		end
	end

	-- make sure the init procedures aren't called again
	for name, module in pairs(_MODULES) do
		if not(_EVENTSTATUS[event][module:GetName()]) then
			_EVENTSTATUS[event][module:GetName()] = true
		end
	end
end

_FRAME.PLAYER_LOGIN = function(self, event, ...)
	if (gCore._ADDONLOADED) then
		_FRAME:ADDON_LOADED("ADDON_LOADED", addon)
	end

	for name, module in pairs(_MODULES) do
		if not(_EVENTSTATUS[event][module:GetName()]) then
			safecall(module.OnEnable, module, ...)
		
			_EVENTSTATUS[event][module:GetName()] = true
		end
	end
end

_FRAME.PLAYER_LOGOUT = function(self, event, ...)
	for name, module in pairs(_MODULES) do
		safecall(module.OnDisable, module, ...)
	end
end

_FRAME.PLAYER_ENTERING_WORLD = function(self, event, ...)
	for name, module in pairs(_MODULES) do
		safecall(module.OnEnter, module, ...)
	end
end

_FRAME:SetScript("OnUpdate", OnUpdate)
_FRAME:SetScript("OnEvent", OnEvent)

------------------------------------------------------------------------------------------------------------
-- 	Template module
-- 	*Serves as a template for both addons and modules 
------------------------------------------------------------------------------------------------------------
--
-- called on ADDON_LOADED
-- use this to initialize values and settings ALL modules need
-- only your "master" module should use this
_MODULE.Init = noop

--
-- called directly after all Init() functions has been called
-- use this to initialize individual modules
_MODULE.OnInit = noop

--
-- called on PLAYER_LOGIN
_MODULE.Enable = noop

--
-- called directly after all Enable() functions has been called
-- use this to create frames and graphics
_MODULE.OnEnable = noop

--
-- called on PLAYER_ENTERING_WORLD
_MODULE.OnEnter = noop

--
-- called on PLAYER_LOGOUT
_MODULE.OnDisable = noop

-- 
-- RegisterEvent(event[, func])
-- @param event <string>
-- @param func <function>
_MODULE.RegisterEvent = function(self, event, func)
	if (type(event) ~= "string") then 
		geterrorhandler()((MAJOR .. ": RegisterEvent(event[, func]) 'event' - string expected, got %s"):format(type(event)), 2)
		return 
	end

	if (func) and (type(func) ~= "function") then
		geterrorhandler()((MAJOR .. ": RegisterEvent(event[, func]) 'func' - function or nil expected, got %s"):format(type(func)), 2)
		return 
	end

	if (_RESERVEDEVENTS[event]) then return end

	if (not _EVENTS[event]) then
		_EVENTS[event] = {}
		_FRAME:RegisterEvent(event)
	end
	
	_EVENTS[event][self:GetName()] = func or event
end

--
-- RegisterEvents([func,] ...)
-- @param func <function>
-- @param ... <string>, <string>, <string>, ...
_MODULE.RegisterEvents = function(self, func, ...)
	if (type(func) == "string") then
		self:RegisterEvent(func)
		func = nil
	end
	
	if (func) and (type(func) ~= "function") then
		error((MAJOR .. ": RegisterEvents([func,] ...) 'func' - function or string expected, got %s"):format(type(func)), 2)
		return 
	end

	for _,event in pairs({...}) do
		if (type(event) ~= "string") then
			error((MAJOR .. ": RegisterEvents([func,] ...) '...' - string expected, got %s"):format(type(event)), 2)
		end
		self:RegisterEvent(event, func)
	end
end

--
-- @param event <string>
_MODULE.UnregisterEvent = function(self, event)
	if (type(event) ~= "string") or (event == "") then return end
	if (_RESERVEDEVENTS[event]) and (not dying) then return end
	
	if (_EVENTS[event][self]) then
		_EVENTS[event][self] = nil
	end
	if (#_EVENTS[event] == 0) then
		_EVENTS[event] = nil
		_FRAME:UnregisterEvent(event)
	end
end

--
-- unregisters ALL events for the given module
_MODULE.UnregisterAllEvents = function(self)
	for event, status in pairs(_RESERVEDEVENTS) do
		self:UnregisterEvent(event)
	end
	
	for event, modules in pairs(_EVENTS) do
		if _EVENTS[event][self:GetName()] then
			_EVENTS[event][self:GetName()] = nil
			self:UnregisterEvent(event)
		end
	end
end

--
-- Schedule a timer for your module
--
-- ScheduleTimer(func[, interval[, delay[, duration]]])
-- @param func <function> the function to be called, the module will be passed as arg #1
-- @param interval <number> interval in seconds. if omitted, the timer will run only once
-- @param delay <number> delay in seconds before the timer starts, optional
-- @param duration <number> indicates how long the function will run in seconds. if omitted, will run until :CancelTimer() is used
-- @return <number> reference to the timer
_MODULE.ScheduleTimer = function(self, func, interval, delay, duration)
	if (type(func) ~= "function") then
		geterrorhandler()((MAJOR .. ": ScheduleTimer(func[, interval[, delay[, duration]]]) 'func' - function expected, got %s"):format(type(func)), 2)
		return 
	end
	
	if (interval) and (type(interval) ~= "number") then
		geterrorhandler()((MAJOR .. ": ScheduleTimer(func[, interval[, delay[, duration]]]) 'interval' - number expected, got %s"):format(type(interval)), 2)
		return 
	end

	if (delay) and (type(delay) ~= "number") then
		geterrorhandler()((MAJOR .. ": ScheduleTimer(func[, interval[, delay[, duration]]]) 'delay' - number expected, got %s"):format(type(delay)), 2)
		return 
	end

	if (duration) and (type(duration) ~= "number") then
		geterrorhandler()((MAJOR .. ": ScheduleTimer(func[, interval[, delay[, duration]]]) 'duration' - number expected, got %s"):format(type(duration)), 2)
		return 
	end

	_TASKS[self:GetName()] = _TASKS[self:GetName()] or {}
	
	local reference = #_TASKS[self:GetName()] + 1

	_TASKS[self:GetName()][reference] = {
		func = func;
		interval = interval;
		delay = delay;
		duration = duration;
	}
	
	return reference
end

--
-- @param reference <number> reference to a scheduled timer
_MODULE.CancelTimer = function(self, reference)
	if (type(reference) ~= "number") then
		geterrorhandler()((MAJOR .. ": CancelTimer(reference) 'reference' - number expected, got %s"):format(type(reference)), 2)
		return 
	end
	
	if (_TASKS[reference]) then
		_TASKS[reference] = nil
	end
end

--
-- @return <string> unique module name
_MODULE.GetName = function(self)
	return tostring(self.NAME) or "nil"
end

--
-- kill a module, unregister its events, cancel its timers, free up its memory
-- does not handle any embeds or child frames or objects
_MODULE.Kill = function(self)
	self.dying = true
	
	self:UnregisterAllEvents()

	_TASKS[self:GetName()] = nil
	_MODULES[self:GetName()] = nil
	
	for i, v in pairs(self) do
		self[i] = nil
	end

	self = nil
end

--
-- CreateChatCommand(func, commands)
-- @param func <function> 
-- @param commands <table> or <string>
_MODULE.CreateChatCommand = function(self, ...)
	gCore.CreateChatCommand(self, ...)
end

--
-- Simple hooking of functions for both secure and unsecure variables
-- There is no checks for previous hooks, so use with caution
-- Main purpose of this function is to simplify the hooking process, 
--  and keeping it to a secure environment whenever possible
-- 
-- Hook(object, method, func)
-- @param object <string, frame> -- for secure hooks, this MUST be a string or a named frame
-- @param method <string> 
-- @param func <function> 
--
-- Hook(object, func)
-- @param object <string> 
-- @param func <function> 
_MODULE.Hook = function(self, object, method, func)
	--
	-- check to see where the function is
	if (not func) then
		func = method
		method = nil
	end
	if (type(func) ~= "function") then
		error((MAJOR .. ": Hook(object[, method], func) 'func' - function expected, got %s"):format(type(func)), 2)
		return 
	end

	local realObject, isFrame, hasHandler, isSecure

	--
	-- Hook(object, method, func) -- hooksecurefunc
	-- @param object <string,table,frame>
	-- @param method <string>
	if (method) then 
		-- 
		-- if there is a method, it must always be a string
		if method and type(method) ~= "string" then
			error((MAJOR .. ": Hook(object, method, func) 'method' - string expected, got %s"):format(type(method)), 2)
			return 
		end
		
		if (type(object) ~= "string") and (type(object) ~= "table") then
			error((MAJOR .. ": Hook(object, method, func) 'object' - string or table (or frame) expected, got %s"):format(type(object)), 2)
			return 
		end
		
		--
		-- do we have an objectname, or an object...?
		if type(object) == "string" then
			if not _G[object] then
				error((MAJOR .. ": Hook(object, method, func) 'object' - no global object named '%s'"):format(object), 2)
				return
			else
				realObject = _G[object]
			end
		elseif type(object) == "table" then
			--
			-- since this is a table, it is most likely the real object
			realObject = object
			
			--
			-- remove the old object, it is not a string, which our functions depend on
			object = nil
			
			if realObject.IsObjectType and realObject:IsObjectType("Frame") then
				isFrame = true
				
				--
				-- if there is a name, put it into the object
				object = realObject:GetName()
			end
		end
		
		--
		-- is the 'method' in fact a script handler?
		hasHandler = realObject.HasScript and realObject:HasScript(method)
	else
		--
		-- Hook(object, func) -- hooksecurefunc
		-- @param object <string> -- name of function to hook
		if type(object) ~= "string" then
			error((MAJOR .. ": Hook(object, func) 'object' - string expected, got %s"):format(type(object)), 2)
			return 
		end
		
		if (not _G[object]) then
			error((MAJOR .. ": Hook(object, func) 'object' - no global function object named '%s'"):format(object), 2)
			return
		else 
			-- since this is to be securely hooked with no 'method', the realObject must be a string
			realObject = object
		end
	end

	--
	-- is it a secure function, table or frame?
	-- this only works on objects entered as strings, or as frames with names
	isSecure = (select(1,(issecurevariable((type(object) == "string") and object or "")))) and (not hasHandler)
	
	--
	-- method is not a script handler, and a secure hook can be used
	if (isSecure) then
		if method then
			if not(realObject[method]) then
				error((MAJOR .. ": Hook(object[, method], func) 'method' - no method named '%s'"):format(method), 2)
				return 
			end
			hooksecurefunc(realObject, method, func)
		else
			hooksecurefunc(realObject, func)
		end
	--
	-- method is a script handler, an unsecure hook must be used
	else
		if isFrame then
			realObject:HookScript(method, func)
		end
	end
end

------------------------------------------------------------------------------------------------------------
-- 	Callback Handler 
------------------------------------------------------------------------------------------------------------
--
-- Register a callback connected to an event
-- if given, func(self, event, ...) will be called, otherwise module[event](self, event, ...) will be called
--
-- RegisterCallback(event[, func])
-- @param event <string> 
-- @param func <function>
-- @return <number> reference to the callback
_MODULE.RegisterCallback = function(self, event, func)
	if (type(event) ~= "string") then
		geterrorhandler()((MAJOR .. ": RegisterCallback(event[, func]) 'event' - string expected, got %s"):format(type(event)), 2)
		return 
	end

	if (func) and (type(func) ~= "function") then
		geterrorhandler()((MAJOR .. ": RegisterCallback(event[, func]) 'func' - function or nil expected, got %s"):format(type(func)), 2)
		return 
	end

	_CALLBACKS[event] = _CALLBACKS[event] or {}
	_CALLBACKS[event][self:GetName()] = _CALLBACKS[event][self:GetName()] or {}

	local reference = #_CALLBACKS[event][self:GetName()]
	
	_CALLBACKS[event][self:GetName()][reference + 1] = func or true

	_CALLBACKFRAME:RegisterEvent(event)
	
	return reference + 1
end


--
-- @param referenceOrEvent <string> event
-- @param referenceOrEvent <number> callback reference
_MODULE.UnregisterCallback = function(self, referenceOrEvent)
	if (type(referenceOrEvent) ~= "number") and (type(referenceOrEvent) ~= "string") then
		geterrorhandler()((MAJOR .. ": UnregisterCallback(referenceOrEvent) 'referenceOrEvent' - string or number expected, got %s"):format(type(referenceOrEvent)), 2)
		return 
	end
	
	if (type(referenceOrEvent) == "string") then
		if _CALLBACKS[referenceOrEvent] then
			if _CALLBACKS[referenceOrEvent][self:GetName()] then
				_CALLBACKS[referenceOrEvent][self:GetName()] = nil
				
				if #_CALLBACKS[referenceOrEvent] == 0 then
					_CALLBACKS[referenceOrEvent] = nil
				end
			end
		end
	elseif (type(referenceOrEvent) == "number") then
		for event, modules in pairs(_CALLBACKS) do
			if _CALLBACKS[event][self:GetName()] and _CALLBACKS[event][self:GetName()][referenceOrEvent] then
				_CALLBACKS[event][self:GetName()][referenceOrEvent] = nil
				
				if #_CALLBACKS[event][self:GetName()] == 0 then
					_CALLBACKS[event][self:GetName()] = nil
				end

				-- removing this for now, as it's not functioning correctly
				-- for some reason it deletes events that still exist :S
--				if #_CALLBACKS[event] == 0 then
--					_CALLBACKS[event] = nil
--					_CALLBACKFRAME:UnregisterEvent(event)
--				end
			end
		end
	end
end

--
-- Unregisters ALL callbacks registered with the given module
_MODULE.UnregisterAllCallbacks = function(self)
	for event, modules in pairs(_CALLBACKS) do
		if _CALLBACKS[event][self:GetName()] then
			self:UnregisterCallback(event)
		end
	end
end

local function OnCallbackEvent(self, event, ...)
	if _CALLBACKS[event] then
		for moduleName, callbacks in pairs(_CALLBACKS[event]) do
			for reference, call in pairs(callbacks) do
				if type(call) == "function" then
					_CALLBACKS[event][moduleName][reference](_MODULES[moduleName], event, ...)
				else
					if _MODULES[moduleName][event] then
						if type(_MODULES[moduleName][event]) ~= "function" then
						else
							_MODULES[moduleName][event](_MODULES[moduleName], event, ...)
						end
					else
						geterrorhandler()(MAJOR .. (": Missing function '%s' in the module named '%s'"):format(event, moduleName))
					end
				end
			end
		end
	end
end
_CALLBACKFRAME:SetScript("OnEvent", OnCallbackEvent)

------------------------------------------------------------------------------------------------------------
-- 	Embed libraries
------------------------------------------------------------------------------------------------------------
--
-- @param (...) <string>, <string>, <string> ...
_MODULE.EmbedLibraries = function(self, ...)
	for i= 1, select("#", ... ) do
		self:EmbedLibrary((select(i, ...)))
	end
end

--
-- @param libname <string>
_MODULE.EmbedLibrary = function(self, libname)
	local lib = LibStub:GetLibrary(libname)
	if (not lib) then
		geterrorhandler()((MAJOR .. ": EmbedLibrary(libname): 'libname' - cannot find a library instance of %q"):format(tostring(libname)), 2)
		
	elseif (lib) and (type(lib.Embed) == "function") then
		lib:Embed(self)
		return true
		
	elseif (lib) then
		error(("Usage: EmbedLibrary(addon, libname, silent, offset): 'libname' - library '%s' is not Embed capable"):format(libname), offset or 2)
	end
end

------------------------------------------------------------------------------------------------------------
-- 	API calls
------------------------------------------------------------------------------------------------------------
local CreateChatCommand = function(self, func, commands)
	if (type(func) ~= "function") then
		geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'func' - function expected, got %s"):format(type(func)), 2)
		return 
	end

	if (type(commands) ~= "table") and (type(commands) ~= "string") then
		geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'commands' - string or table expected, got %s"):format(type(commands)), 2)
		return 
	end

	local name = MAJOR .. "_ChatCommand" .. tostring(#_SLASH + 1)
	tinsert(_SLASH, name)
	
	_G["SlashCmdList"][name] = function(argString) 
		argString = gsub(argString, "[%s]+", " ")
		func(strsplit(" ", argString))
	end

	if (type(commands) == "table") then 
		for i,v in pairs(commands) do
			if (type(v) ~= "string") then
				geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'commands' - values in this table must be strings, got %s at commands[%s]"):format(type(v), tostring(i)), 2)
				return 
			end

			_G["SLASH_" .. name .. i] = "/" .. v
		end
	elseif (type(commands) == "string") then 
		_G["SLASH_" .. name .. "1"] = "/".. commands
	end
end

--
-- @param oldClass <table>
-- @return <table> which inherits from oldClass
local NewClass = function(oldClass)
    local newClass = {
		New = function()
			return setmetatable({}, { __index = newClass })
		end
	}

    if (oldClass) then
        setmetatable(newClass, { __index = oldClass })
    end
	
    return newClass
end

--
-- @param name <string> 
-- @param parent <table> object to inherit from
-- @param (...) <string>, <string>, <string> ... list of libraries to embed
-- @return <table> 
local NewObject = function(name, parent, ...)
	local new = NewClass(parent)
	new._NAME = name

	if (select("#", ...) > 0) then
		new:EmbedLibraries(...)
	end
	
	return new
end

--
-- @param name <string> unique module name (not a global)
-- @param (...) <string>, <string>, <string> ... list of libraries to embed
-- @return <table> new module
local NewModule = function(self, name, ...)
	if (type(name) ~= "string") then 
		geterrorhandler()((MAJOR .. ": NewModule(name): 'name' - string expected, got %s"):format(type(name)), 2)
		return 
	end

	if (self._MODULES[name]) then
		geterrorhandler()((MAJOR .. ": NewModule(name): 'name' - a module named '%s' already exists"):format(name), 2)
		return
	end

	self._MODULES[name] = NewObject(self)

	return self._MODULES[name]
end

--
-- @param name <string> unique module name
-- @return <table> module[name] or nil
local GetModule = function(self, name)
	if not name or type(name) ~= "string" or name == "" or not(self._MODULES[name]) then return end
	
	return self._MODULES[name]
end

--
-- @param name <string> unique addon name (not a global)
-- @param (...) <string>, <string>, <string> ... list of libraries to embed
-- @return <table> new module
gCore.NewAddon = function(self, name, ...)
	if (type(name) ~= "string") then 
		geterrorhandler()((MAJOR .. ": NewAddon(name): 'name' - string expected, got %s"):format(type(name)), 2)
		return 
	end

	if (_ADDONS[name]) then
		geterrorhandler()((MAJOR .. ": NewAddon(name): 'name' - a module named '%s' already exists"):format(name), 2)
		return
	end
	
	_ADDONS[name] = NewObject(name, _MODULE, ...)
	_ADDONS[name]._MODULES = {}
	_ADDONS[name]._CALLBACKS = {}
	_ADDONS[name].NewModule = NewModule
	_ADDONS[name].GetModule = GetModule

	return _ADDONS[name]
end

--
-- @param name <string> unique module name (not a global)
-- @param (...) <string>, <string>, <string> ... list of library names to embed
-- @return <table> new module
function gCore:NewModule(name, ...)
	if type(name) ~= "string" then 
		geterrorhandler()((MAJOR .. ": NewModule(name): 'name' - string expected, got %s"):format(type(name)), 2)
		return 
	end

	if _MODULES[name] then
		geterrorhandler()((MAJOR .. ": NewModule(name): 'name' - a module named '%s' already exists"):format(name), 2)
		return
	end
	
	_MODULES[name] = NewClass(_MODULE)
	_MODULES[name].NAME = name
	
	if select("#", ...) > 0 then
		_MODULES[name]:EmbedLibraries(...)
	end
	
	return _MODULES[name]
end

--
-- @param name <string> unique module name
-- @return <table> module[name] or nil
function gCore:GetModule(name)
	if not name or type(name) ~= "string" or name == "" or not(_MODULES[name]) then return end
	
	return _MODULES[name]
end

function gCore:GetAllModules()
	return _MODULES
end

--
-- @param func <function> 
-- @param commands <table> or <string>
function gCore:CreateChatCommand(func, commands)
	if type(func) ~= "function" then
		geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'func' - function expected, got %s"):format(type(func)), 2)
		return 
	end

	if type(commands) ~= "table" and type(commands) ~= "string" then
		geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'commands' - string or table expected, got %s"):format(type(commands)), 2)
		return 
	end

	local name = MAJOR .. "_ChatCommand" .. tostring(#_SLASH + 1)
	tinsert(_SLASH, name)
	
	_G["SlashCmdList"][name] = function(argString) 
		func(strsplit(" ", argString))
	end

	if type(commands) == "table" then 
		for i,v in pairs(commands) do
			if type(v) ~= "string" then
				geterrorhandler()((MAJOR .. ": CreateChatCommand(func, commands) 'commands' - values in this table must be strings, got %s at commands[%s]"):format(type(v), tostring(i)), 2)
				return 
			end

			_G["SLASH_" .. name .. i] = "/" .. v
		end
	elseif type(commands) == "string" then 
		_G["SLASH_" .. name .. "1"] = "/".. commands
	end
end

------------------------------------------------------------------------------------------------------------
-- 	Debugging
------------------------------------------------------------------------------------------------------------
function gCore:list()
	print("|cffffffffModules:|r")
	for name, module in pairs(_MODULES) do
		print((" |cffffaa00%s|r"):format(name))
	end
end

------------------------------------------------------------------------------------------------------------
-- 	Global functions
------------------------------------------------------------------------------------------------------------
--
-- set multiple CVars at once
--
-- @param cvars <table> { [CVarName] = <string> value }
function SetCVars(cvars)
	for i,v in pairs(cvars) do
		safecall(SetCVar(i, v))
	end
end

--
-- returns 1 if 'choice' equals true or 1, and 0 for false, 0 or nil
function GetBinary (choice)
	return ((choice == 1 or choice == true) and 1) or (not choice or choice == 0) and 0 or 0
end

--
-- returns true if 'choice' equals true or 1, or false if it equals false, 0 or nil
--
function GetBoolean(choice)
	return ((choice == true or choice == 1) and true) or ((not choice or choice == 0) and false) or false
end

--
-- embeds every key from source into target
--
-- @param target <table> target table
-- @param source <table> source table
-- @return <table> target table
function Embed(target, source)
	if not target or not source then return end
	if type(source) ~= "table" or type(target) ~= "table" then return end 

	for i,v in pairs(source) do
		target[i] = v
	end
	
	return target
end

--
-- deletes entries in the target table that aren't found in source
-- does NOT iterate through subtables! (which is intended)
--
-- @param target <table> target table
-- @param source <table> source table
-- @return <table> target table
function CleanTable(target, source)
	if not source or type(source) ~= "table" or not target or type(target) ~= "table" then return end
	
	for i,_ in pairs(target) do
		if source[i] == nil then
			target[i] = nil
		end
	end
	
	return target
end

--
-- fills in the holes in the target table with values from source
-- will clean out unknown keys unless 'noClean' is set to 'true'
--
-- this function is mainly intended for validating saved settings, 
-- and killing off deprecated entries
--
-- ValidateTable(target[, source[, noClean]]) 
-- @param target <table> target table
-- @param source <table> source table
-- @param noClean <boolean>
-- @return <table> target table
function ValidateTable(target, source, noClean) 
	if not source or type(source) ~= "table" then return end

	if not target or type(target) ~= "table" then 
		target = {}
	end

	for i,_ in pairs(source) do
		if target[i] == nil or type(target[i]) ~= type(source[i]) then
			if type(source[i]) == "table" then
				target[i] = CopyTable(source[i])
			else
				target[i] = source[i]
			end
		elseif type(target[i]) == "table" then
			target[i] = ValidateTable(target[i], source[i])
		end
	end
	
	if not noClean then
		CleanTable(target, source)
	end
	
	return target
end

--
-- not happy with these, they should be embedded in the library and/or MODULE instead
--
_G.Embed = Embed
_G.GetBinary = GetBinary
_G.GetBoolean = GetBoolean
_G.SetCVars = SetCVars
_G.CleanTable = CleanTable
_G.ValidateTable = ValidateTable
