﻿local addon, ns = ...

local MAJOR, MINOR = "gFrameHandler-1.0", 2
local gFrameHandler, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

-- gFrameHandler already in place
if not gFrameHandler then return end 

-- Lua API
local type, tonumber = type, tonumber
local print, gsub = print, gsub
local floor = math.floor
local tinsert = table.insert
local pairs, unpack = pairs, unpack
local setmetatable = setmetatable

local eventHandler = CreateFrame("Frame")
eventHandler:RegisterEvent("VARIABLES_LOADED")
eventHandler:RegisterEvent("PLAYER_LOGIN")
eventHandler:RegisterEvent("PLAYER_ENTERING_WORLD")
eventHandler:RegisterEvent("PLAYER_REGEN_DISABLED")
eventHandler:SetScript("OnEvent", function(self, event, ...) 
	gFrameHandler[event](gFrameHandler, event, ...)
end)

gFrameHandler.Objects = gFrameHandler.Objects or {}
gFrameHandler.Queue = gFrameHandler.Queue or {}
gFrameHandler.Positions = gFrameHandler.Positions or {}
gFrameHandler.Anchors = gFrameHandler.Anchors or {}

local print = function(...)
	return _G.print((GetAddOnMetadata(addon, "Title") or addon) .. ":" , ...)
end

local round = function(n)
	return floor(n * 1e5 + .5) / 1e5
end

------------------------------------------------------------------------------------------------------------
-- 	Extended/Improved Frame- And Texture Handling
------------------------------------------------------------------------------------------------------------

--
-- this function will always return the position relative to UIParent
local function GetPoint(object)
	local UIcenterX, UIcenterY = UIParent:GetCenter()
	local objectX, objectY = object:GetCenter()

	if not objectX then return end

	local UIScale = UIParent:GetEffectiveScale()
	local objectScale = object:GetEffectiveScale()

	local UIWidth, UIHeight = UIParent:GetRight(), UIParent:GetTop()

	local LEFT = UIWidth / 3
	local RIGHT = UIWidth * 2 / 3
	local BOTTOM = UIHeight / 3
	local TOP = UIHeight * 2 / 3

	local point, x, y
	if objectX >= RIGHT then
		point = "RIGHT"
		x = object:GetRight() - UIWidth
	elseif objectX <= LEFT then
		point = "LEFT"
		x = object:GetLeft()
	else
		x = objectX - UIcenterX
	end

	if objectY >= TOP then
		point = "TOP" .. (point or "")
		y = object:GetTop() - UIHeight
	elseif objectY <= BOTTOM then
		point = "BOTTOM" .. (point or "")
		y = object:GetBottom()
	else
		if not point then point = "CENTER" end
		y = objectY - UIcenterY
	end

	return point, "UIParent", point, round(x * UIScale / objectScale),  round(y * UIScale / objectScale)
end

------------------------------------------------------------------------------------------------------------
-- 	Saved position handlers
------------------------------------------------------------------------------------------------------------
local function SetSavedPosition(object)
	if object:GetName() then
		gFrameHandler.Positions[object:GetName()] = {GetPoint(object)}
	end
end

local function GetSavedPosition(object)
	if not object:GetName() or not gFrameHandler.Positions[object:GetName()] then return end

	return unpack(gFrameHandler.Positions[object:GetName()])
end

local function DeleteSavedPosition(object)
	if not object:GetName() or not gFrameHandler.Positions[object:GetName()] then return end
	
	gFrameHandler.Positions[object:GetName()] = nil
end

local function RestoreToSavedPosition(object)
	if not object:GetName() or not gFrameHandler.Positions[object:GetName()] then return end
	
	if not InCombatLockdown() then
		local point, anchor, relpoint, x, y = GetSavedPosition(object)
		if point then
			object:ClearAllPoints()
			object:SetPoint(point, anchor, relpoint, x, y)
		end
	end
end

--
-- Attempting to remove common GUIS prefixes to make the display names more readable
-- Will have to work a bit on this one
--
-- My search patterns suck, I really need to update my skills there a bit
local function GetSmartName(object)
	if not object.GetName or not object:GetName() or object:GetName() == "" then return "" end
	
	local name = object:GetName()
	
	name = gsub(name, "(GUIS)", "")
	name = gsub(name, "(gUI)", "")
	name = gsub(name, "(: )", "")
	name = gsub(name, "(:)", "")
	name = gsub(name, "(-)", " ")
	name = gsub(name, "(_)", " ")
	
	while strsub(name, 1, 1) == " " do
		name = strsub(name, 2)
	end
	
	name = gsub(name, "(%l)(%u)", function(l, u) return l .. " " .. u end)
	
	return name
end

local dragTexture = {0.8, 0.8, 1, 0.5}
local hitRectInsets = {-15, -15, -15, -15}
local function CreateAnchor(object)
	if gFrameHandler.Anchors[object:GetName()] and gFrameHandler.Anchors[object:GetName()].anchor then return end
	gFrameHandler.Anchors[object:GetName()] = gFrameHandler.Anchors[object:GetName()] or {}
	
	local anchor = CreateFrame("Frame", nil, UIParent)
	anchor:Hide()
	anchor:SetFrameStrata("DIALOG")
	anchor:SetSize(object:GetSize())
	anchor:SetPoint(GetPoint(object))
	anchor:SetBackdrop({ bgFile = "Interface\\ChatFrame\\ChatFrameBackground" })
	anchor:SetBackdropColor(1, 1, 1, 1/3)
	anchor:SetHitRectInsets(-15, -15, -15, -15)
	anchor:EnableMouse(true)
	anchor:SetMovable(true)
	anchor:SetUserPlaced(true)
	anchor:RegisterForDrag("LeftButton", "RightButton")

	anchor.isLocked = true
	anchor.frame = object

	if not gFrameHandler.Anchors[object:GetName()].defaultPosition then 
		gFrameHandler.Anchors[object:GetName()].defaultPosition = {GetPoint(object)}
	end
	
	local anchorTitle = anchor:CreateFontString(nil, "OVERLAY")
	anchorTitle:SetPoint("CENTER", anchor, "CENTER")
	anchorTitle:SetFontObject(SystemFont_Outline)
	anchorTitle:SetText(GetSmartName(object))
	anchorTitle:SetTextColor(1, 1, 1, 1)

	anchor:SetScript("OnDragStart", function(self)
		self:StartMoving()

		self.oldAlpha = self.frame:GetAlpha()

		self.frame:SetAlpha(0)
	end)

	anchor:SetScript("OnDragStop", function(self) 
		self:StopMovingOrSizing()

		self.frame:ClearAllPoints()
		self.frame:SetPoint(GetPoint(self))
		self.frame:SetAlpha(self.oldAlpha)

		self.oldAlpha = nil
		
		SetSavedPosition(self.frame)
	end)

	gFrameHandler.Anchors[object:GetName()].anchor = anchor
end

local function Lock(object)
	if not(gFrameHandler.Anchors[object:GetName()] and gFrameHandler.Anchors[object:GetName()].anchor) then return end

	gFrameHandler.Anchors[object:GetName()].anchor.isLocked = true
	gFrameHandler.Anchors[object:GetName()].anchor:Hide()
end

local function Unlock(object)
	if not(gFrameHandler.Anchors[object:GetName()] and gFrameHandler.Anchors[object:GetName()].anchor) then 
		CreateAnchor(object)
	end

	if not InCombatLockdown() then
		gFrameHandler.Anchors[object:GetName()].anchor.isLocked = false
		gFrameHandler.Anchors[object:GetName()].anchor:SetSize(object:GetSize())
		gFrameHandler.Anchors[object:GetName()].anchor:ClearAllPoints()
		gFrameHandler.Anchors[object:GetName()].anchor:SetPoint(GetPoint(object))
		gFrameHandler.Anchors[object:GetName()].anchor:Show()
	end
end

local function Reset(object)
	if not object.GetName or not gFrameHandler.Positions[object:GetName()] or not gFrameHandler.Objects[object:GetName()] then return end
	
	if not InCombatLockdown() then
		object:ClearAllPoints()
		object:SetPoint(unpack(gFrameHandler.Anchors[object:GetName()].defaultPosition))
		DeleteSavedPosition(object)
	end
end

local function IsLocked(object)
	return (not(gFrameHandler.Anchors[object:GetName()]) or (gFrameHandler.Anchors[object:GetName()] and (gFrameHandler.Anchors[object:GetName()].anchor.isLocked == true))) and true or false
end

local function RegisterForSave(object, ...)
	if not object:GetName() or object:GetName() == "" then return end

	gFrameHandler.Anchors[object:GetName()] = gFrameHandler.Anchors[object:GetName()] or {}
	gFrameHandler.Objects[object:GetName()] = gFrameHandler.Objects[object:GetName()] or {}
	gFrameHandler.Objects[object:GetName()].object = object
	
	if (...) then
		gFrameHandler.Anchors[object:GetName()].defaultPosition = {...}
	else
		if IsLoggedIn() then
			gFrameHandler.Anchors[object:GetName()].defaultPosition = {GetPoint(object)}
		else
			tinsert(gFrameHandler.Queue, object:GetName())
		end
	end
end

local RestoreAllSavedPositions
do
	function RestoreAllSavedPositions()
		if not InCombatLockdown() then
			for object, position in pairs(gFrameHandler.Positions) do
				if _G[object] and _G[object].SetPoint then
					Lock(_G[object])
					_G[object]:ClearAllPoints()
					_G[object]:SetPoint(unpack(position))
				end
			end
		else
			eventHandler:RegisterEvent("PLAYER_REGEN_ENABLED")
		end
	end
end

local function ClearAllSavedPositions()
	wipe(gFrameHandler.Positions)
end

------------------------------------------------------------------------------------------------------------
-- 	Add the library API
------------------------------------------------------------------------------------------------------------
function gFrameHandler:GetPoint(object)
	return GetPoint(object)
end

function gFrameHandler:Register(object, ...)
	if not object or not object.IsObjectType or not object:IsObjectType("Frame") or not object:GetName() or object:GetName() == "" then 
		return 
	end

	RegisterForSave(object, ...)
end

function gFrameHandler:Unregister(object)
	if not object or not object.GetName or not object:GetName() or object:GetName() == "" then return end
	
	self.Objects[object:GetName()] = nil
	self.Anchors[object:GetName()] = nil

	DeleteSavedPosition(object)
end

local _UNLOCK
function gFrameHandler:Unlock()
	if not InCombatLockdown() then
		_UNLOCK = true
		for _,object in pairs(self.Objects) do
			Unlock(object.object)
		end
	end
end

function gFrameHandler:Lock()
	_UNLOCK = nil
	for _,object in pairs(self.Objects) do
		Lock(object.object)
	end
end

function gFrameHandler:Reset(bar)
	self:Lock()
	
	if not(InCombatLockdown()) then
		if (bar) then
			if (bar.GetObjectType) and (bar:IsObjectType("Frame")) and (self.Anchors[bar:GetName()]) then
				bar:ClearAllPoints()
				bar:SetPoint(unpack(self.Anchors[bar:GetName()].defaultPosition))			
				
				DeleteSavedPosition(bar)
				
			elseif (type(bar) == "table") then
				for name, object in pairs(bar) do
					if (object.GetObjectType) and (object:IsObjectType("Frame")) and (self.Anchors[object:GetName()]) then
						object:ClearAllPoints()
						object:SetPoint(unpack(self.Anchors[object:GetName()].defaultPosition))				
					
						DeleteSavedPosition(object)
					end
				end
			end
		else
			for name, object in pairs(self.Objects) do
				object.object:ClearAllPoints()
				object.object:SetPoint(unpack(self.Anchors[object.object:GetName()].defaultPosition))
			end
			
			ClearAllSavedPositions()
		end
	end
end

function gFrameHandler:ToggleLock()
	if _UNLOCK then
		self:Lock()
	else
		self:Unlock()
	end
end

function gFrameHandler:SetSavedPositions(tbl)
	if type(tbl) ~= "table" then
		geterrorhandler()((MAJOR .. ": SetSavedPositions(tbl) 'tbl' - table expected, got %s"):format(type(tbl)), 2)
		return 
	end

	self.Positions = tbl
end

function gFrameHandler:GetSavedPositions()
	return gFrameHandler.Positions
end

function gFrameHandler:PLAYER_LOGIN()
	for _,fName in pairs(self.Queue) do
		if _G[fName] then
			self.Anchors[fName].defaultPosition = {GetPoint(_G[fName])}
		end
	end
end

function gFrameHandler:PLAYER_ENTERING_WORLD()
	RestoreAllSavedPositions()
end

function gFrameHandler:VARIABLES_LOADED()
	RestoreAllSavedPositions()
end

function gFrameHandler:PLAYER_REGEN_DISABLED()
	if _UNLOCK then
		self:Lock()
	end
end

function gFrameHandler:PLAYER_REGEN_ENABLED()
	eventHandler:UnregisterEvent("PLAYER_REGEN_ENABLED")
	
	RestoreAllSavedPositions()
end

------------------------------------------------------------------------------------------------------------
-- 	API calls
------------------------------------------------------------------------------------------------------------
do
	local function setAPI(frame)
		local meta = getmetatable(frame).__index

		if not meta.RegisterForSave then meta.RegisterForSave = RegisterForSave end
	end

	local done = {}

	local frames = EnumerateFrames()
	while frames do
		if not done[frames:GetObjectType()] then
			setAPI(frames)
			done[frames:GetObjectType()] = true
		end

		frames = EnumerateFrames(frames)
	end
	
	done = nil
end
